/*
 * Timer1 is setup as a center aligned PWM with complimentary outputs and 560ns dead time. It will generate an ISR when the overflow value is reached. All PLL code will run in this interrupt
 * The PLL uses a 1/2 park transform to act as a frequency / phase detector. When V_grid and the reference DCO(Digital Controlled Oscillator) are 90º out of phase 
 * the q output of the park transform will be zero. q is the phase error which is fed to the input of a PI controller
 * The output of the PI controler is mapped over to be the input overflow value for timer1. This closes the loop locking the grid signal and DCO output in quadrature
 * The base squarewave on PB4 and the pwm outputs on PA8,PB13 & PA9,PB14 are in sync with the grid thanks to software counter sinInc_90 which runs +90º relative to counter sinInc
 * The pwm signal is provided on PA8,PB13 & PA9,PB14 and can either be a pure 120Hz Halfsine or it can follow the grid voltage input on PA0
 * Data capture and display is available and will provide one full cycle of data on three variables after a button press as not to disturb the PLL.
 * Calculated constants are based on the user selected constants. Table one shows the effects of different increments on the PLL
 * The default increments is 360 for a good balance between Phase adjust and phase jitter while maintaing a good pwm frequency of 21.6kHz
 */
#include <STM32ADC.h>
STM32ADC myADC(ADC1);
#include <JC_Button.h>// https://github.com/JChristensen/JC_Button
Button PUSH_BUTTON(PB1, 150);
/* Table1
 * Increments | pwm_ovf value | PWM Frequency |   ISR    | Phase Adj Increment  | Phase Jitter
 *    256     |     2343      |   15.365kHz   | 65.08us  |      +/-1.4º         |   +/-0.15º
 *    360     |     1666      |   21.609kHz   | 46.28us  |      +/-1º           |   +/-0.22º
 *    400     |     1500      |   24.000KHZ   | 41.67us  |      +/-0.9º         |   +/-0.24º
 *    512     |     1171      |   30.743kHz   | 32.5us   |      +/-0.7º         |   +/-0.3º
 */
// User selected constants
const int clk_freq = 72E06;// Microcontroller nominal clock frequency
const int incrmnts = 360;// Number of increments for a 360º sine-wave (must be evenly divisable by 2)
const float grid_freq = 60.0;// Nominal grid frequency

/* The phase may need to be compensated for transformer and other circuit effects. This is accomplised by adding or subtracting from the +90º offset
   and will help bring the reference DCO closer in phase with the grid.
   Table1 shows the minimum phase adjust increment based on the total number of full sinewave increments
*/
const float phase_adjust = 2.0;// Phase adjust increment

/*Calculated constants*/
const float grid_freq_upper = grid_freq + .5;// Grid frequency upper bound for pll lock
const float grid_freq_lower = grid_freq - .5;// Grid frequency lower bound for pll lock
const int phase_correction = incrmnts / 4;// corresponds to a 90º phase shift
const int half_incrmnts = incrmnts / 2;
const float grid_per = 1/60.;
const float incrmnt_sec = grid_per/incrmnts;
const int isr_freq = 1/incrmnt_sec;
const int pwm_ovf_edge = clk_freq / isr_freq;// Nominal overflow value for PWM edge aligned mode
const int pwm_ovf_center = (clk_freq / isr_freq)/2;// Nominal overflow value for PWM center aligned mode
const int pwm_ovf_upper = (grid_freq_upper / grid_freq)*pwm_ovf_center;
const int pwm_ovf_lower = (grid_freq_lower / grid_freq)*pwm_ovf_center;

// Global variables
static int DCO_sinArry[incrmnts];//array containg the DCO sine wave
static float cos_theta[incrmnts];//0 +/- 1.0 cosine wave
static float sin_theta[incrmnts];//0 +/- 1.0 sine wave
int pwm_ovf = pwm_ovf_center;// Timer1 overflow value 
int V_grid;// Grid sample
volatile int sinInc;// Software counter 1
volatile int sinInc_90 = phase_correction + phase_adjust;// Software counter 2
int DCO_sine;
int pwmVal;
volatile float mod_factor = 1.0;// Controls the amplitude of the pwm output
float cos_angle;
float sin_angle;

// Data capture and display arrays
int data_arry1[incrmnts];
int data_arry2[incrmnts];
int data_arry3[incrmnts];
int data_arry4[incrmnts];
//***************************

// PLL PI Tuning variables
//**************************
int pll_out;
int phase_error;
int q;
float PI_p, PI_i;
float Kp = 10, Ki = .05;
//**************************

void setup() {
  Serial.begin(115200);
  //calibrate ADC.
  myADC.calibrate(); 
  //myADC.setSampleRate(ADC_SMPR_7_5); // approx 3us per sample
  pinMode(PA0, INPUT_ANALOG);// Grid voltage
  pinMode(PB1, INPUT_PULLUP);//BUTTON
  pinMode(PB3, OUTPUT);// TEST1
  pinMode(PB4, OUTPUT);// 60Hz sq-wave scope probe signal
  pinMode(PB5, OUTPUT);// TEST2
  
  delay(250);
  // Set up  Timer1 for Complimentary PWM on Channel 1 and Channel 2
  Timer1.setPrescaleFactor(1);//1 to 65,536 Freq = (72MHz / Prescale_val=1)
  Timer1.setOverflow(pwm_ovf);
  bitSet(TIMER1_BASE->CCER,2); //Sets CC1NE bit enable complimentary Channel l Output
  bitSet(TIMER1_BASE->CCER,6); //Sets CC2NE bit enable complimentary Channel 2 Output
  //Inserts Dead Time approx 14ns per increment 0 to 128
  //(DTG Bits 7:0)
  //DGT Bit 2&5 = 00100100 = 560ns
  bitSet(TIMER1_BASE->BDTR,2);
  bitSet(TIMER1_BASE->BDTR,5);
    /* Reference Manual p338 - Set CMS bit6 for Center-aligned mode 2. The counter counts up and down alternatively. Output compare
   interrupt flags of channels configured in output (CCxS=00 in TIMx_CCMRx register) are set
   only when the counter is counting up*/
  bitSet(TIMER1_BASE->CR1,6);

  // When CH1 & CH2 are set to outputs then use CH3 or CH4 to get the interrupt
  Timer1.attachInterrupt(TIMER_CH3, TIM1_ISR);
  
  // Initially Shut OFF pwm
  pwmWrite(PA8,0);
  pwmWrite(PA9,0);

  delay(250);
  
  for(int i = 0; i < incrmnts; i++)
  { 
    // Create sine array to near the nominal peak value of the grid sample
    int val = sin(i*M_PI/(half_incrmnts)) * 1700;
    DCO_sinArry[i] = val;
    // Create full sine and cosine arrays -1 to +1
    float val2 = sin(i * M_PI/half_incrmnts);
    sin_theta[i] = val2;
    float val3 = cos(i * M_PI/half_incrmnts);
    cos_theta[i] = val3;
  }
  
  // Wait 2 sec before turing on the PWM GPIO's
  // This will prevent output transients since PA8 and PA9 have already been written to zero during Timer1 setup
  delay(1000);
  pinMode(PA8, PWM);//Timer1 Channel 1(T1C1) output assigned to PA8 and setting it to PWM Output (Q1)
  pinMode(PB13, PWM);//Timer1 Channel 1(T1C1N) complimentary output assigned to PB13 and setting it to PWM Output(Q3)
  pinMode(PA9, PWM);//Timer1 Channel 2(T1C2) output assigned to PA9 and setting it to PWM Output(Q2)
  pinMode(PB14, PWM);//Timer1 Channel 2(T1C2N) complimentary output assigned to PB14 and setting it to PWM Output(Q4)

} //end setup

void TIM1_ISR(){
    gpio_write_bit(GPIOB,3,1);// ISR execution time = 22.2us
    V_grid = analogRead(PA0)-2036;// Compensate for resistor divider error and calibrate to zero by turning off grid voltage and display only v_grid
    
    //Data Capture
    data_arry1[sinInc_90] = gpio_read_bit(GPIOB,4)* 12;// Add an additional variable to display
    data_arry2[sinInc_90] = V_grid * .1;// Normalize for scaling with phase_error
    data_arry3[sinInc] = DCO_sine * .1;// Normalize for scaling with phase_error 
    data_arry4[sinInc_90] = phase_error;

    if(++sinInc >= incrmnts)
    {
      sinInc = 0;//
    } 
    if(++sinInc_90 >= incrmnts)
    {
      sinInc_90 = 0;//
    } 
    
    // PLL
    // Timer1 interupts every 46.28us this corresponds to 60Hz grid taking 360 samples per cycle(46.28us * 360 incrmnts)
    // Variable "sinInc" is advanced by one count for each interrupt up to 359 these counts represent a φ of 1º per cnt
    // Using 1/2 of a park transformation as a frequency/phase detector where as q = (ß*cosφ)-(α * sinφ) where ß= internal oscillator DCO_sinArry[] and α = V_grid
    // cosφ and sinφ are generated in setup() and each are stored in an array.
    // The PI controller keeps Timer1 in sync with the grid by adjusting the timer1 overflow value every increment (1º) and has a max phase jitter of +/- 0.22º
    // maintaing a 90º difference between α and ß resulting in a q = phase error of 0 and a more positve error when grid freq is higher and visa-versa

    // Frequency / Phase Discriminator (1/2 Park Transformation)
    sin_angle = sin_theta[sinInc];
    cos_angle = cos_theta[sinInc];
    DCO_sine = DCO_sinArry[sinInc];
    q = (DCO_sine * cos_angle) - (V_grid * sin_angle);//
    phase_error = -q;
    // PI Controller
    PI_p = Kp * phase_error;
    PI_i = PI_i + (Ki * phase_error);
    pll_out =PI_p + PI_i;
    pll_out = map(pll_out, -1000,1000, pwm_ovf_lower, pwm_ovf_upper);
    pwm_ovf = pll_out;
    
    Timer1.setOverflow(pwm_ovf);


    // Timer1 PWM value is either pure reference sinewave or grid sinewave
    // The mod_factor can control the amplitude of the pwm
    //pwmVal = mod_factor * abs(V_grid);
    pwmVal = mod_factor * abs(DCO_sinArry[sinInc_90]);

    
    // Generate a base square-wave and pwm in sync with the grid      
    if(sinInc_90 < half_incrmnts)
    {
      pwmWrite(PA8, pwmVal);
      pwmWrite(PA9, 0);
      gpio_write_bit(GPIOB,4,1);
    }
    else
    {
      pwmWrite(PA9, pwmVal);
      pwmWrite(PA8, 0);      
      gpio_write_bit(GPIOB,4,0);
    }  
  gpio_write_bit(GPIOB,3,0); 
}

void loop(){
    PUSH_BUTTON.read();
  if (PUSH_BUTTON.wasReleased())
    {
      data_capture_dsply();
    }

} //end loop

void data_capture_dsply(){

      for(int i=0; i < incrmnts; i++)
      {
        Serial.print("PB4:");// Add an additional variable to display and uncomment this line and the next two lines
        Serial.print(data_arry1[i]);
        Serial.print(",");        
        Serial.print("V_grid:");
        Serial.print(data_arry2[i]);
        Serial.print(",");
        Serial.print("DCO_Sine:");
        Serial.print(data_arry3[i]);
        Serial.print(",");        
        Serial.print("Phase_Error:");
        Serial.println(data_arry4[i]);
      }

}
